import json
from texttable import Texttable

from nexuscli import exception
from nexuscli.api import repository
from nexuscli.cli import constants


def cmd_list(nexus_client):
    """Performs ``nexus3 repository list``"""
    repositories = nexus_client.repositories.raw_list()

    table = Texttable(max_width=constants.TTY_MAX_WIDTH)
    table.add_row(['Name', 'Format', 'Type', 'URL'])
    table.set_deco(Texttable.HEADER)
    for repo in repositories:
        table.add_row(
            [repo['name'], repo['format'], repo['type'], repo['url']])

    print(table.draw())
    return exception.CliReturnCode.SUCCESS.value


def cmd_create(ctx, repo_type=None, repository_name=None, **kwargs):
    """Performs ``nexus3 repository create`` commands"""
    nexus_client = ctx.obj
    recipe = kwargs["recipe"]
    kwargs['nexus_client'] = nexus_client

    Repository = repository.collection.get_repository_class({
        'recipeName': f'{recipe}-{repo_type}'})

    r = Repository(repository_name, **kwargs)

    nexus_client.repositories.create(r)

    return exception.CliReturnCode.SUCCESS.value


def cmd_delete(nexus_client, repository_name):
    """Performs ``nexus3 repository delete``"""
    nexus_client.repositories.delete(repository_name)
    return exception.CliReturnCode.SUCCESS.value


def cmd_show(nexus_client, repository_name):
    """Performs ``nexus3 repository show"""
    repo_name = repository_name
    try:
        configuration = nexus_client.repositories.get_raw_by_name(repo_name)
    except exception.NexusClientInvalidRepository:
        print(f'Repository not found: {repo_name}')
        return exception.CliReturnCode.REPOSITORY_NOT_FOUND.value

    print(json.dumps(configuration, indent=2))
    return exception.CliReturnCode.SUCCESS.value



# -*- coding: utf-8 -*-
import hashlib
import mmap
import os
import pathlib
import pkg_resources


def _resource_filename(resource_name):
    """wrapper for pkg_resources.resource_filename"""
    return pkg_resources.resource_filename('nexuscli', resource_name)


def groovy_script(script_name):
    """
    Returns the content for a groovy script located in the package installation
    path under script/groovy.

    E.g.: groovy_script('foo') returns the content for the file at
    ``.../site-packages/nexuscli/script/groovy/foo.groovy``.

    :param script_name: file name of the groovy script; ``.groovy`` is appended
        to this parameter to form the file name.
    :return: content for the groovy script
    :rtype: str
    """
    script_path = os.path.join(
        'api', 'script', 'groovy', f'{script_name}.groovy')
    script_path = _resource_filename(script_path)
    return open(script_path).read()


def validate_strings(*args):
    """
    Checks that all given arguments have a string type (e.g. str, basestring,
    unicode etc)

    Args:
        *args: values to be validated.

    Returns:
        bool: True if all arguments are of a string type. False otherwise.
    """
    for arg in args:
        if not isinstance(arg, str):
            return False

    return True


def filtered_list_gen(raw_response, term=None, partial_match=True):
    """
    Iterates over items yielded by raw_response_gen, validating that:
        1. the `path` dict key is a str
        2. the `path` value starts with starts_with (if provided)

    >>> r = [{
    >>>     'checksum': {
    >>>         'md5': 'd94b865aa7620c46ef8faef7059a311c',
    >>>         'sha1': '2186934d880cf24dd9ecc578335e290026695522',
    >>>         'sha256': 'b7bb3424a6a6(...)4113bc38fd7807528481a8ffe3cf',
    >>>         'sha512': 'e7806f3caa3e(...)3caeb9bbc54bbde286c07f837fdc'
    >>>     },
    >>>     'downloadUrl': 'http://nexus/repository/repo_name/a/file.ext',
    >>>     'format': 'yum',
    >>>     'id': 'Y2xvdWRlcmEtbWFuYWdlcj(...)mRiNWU0YjllZWQzMg',
    >>>     'path': 'a/fake.rpm',
    >>>     'repository': 'cloudera-manager'}]
    >>>
    >>> for i in filtered_list_gen(r, starts_with='a/fake.rpm')
    >>>     print(i['path'])
    a/fake.rpm
    >>> for i in filtered_list_gen(r, starts_with='b')
    >>>     print(i['path'])
    # (nothing printed)

    Args:
        raw_response (iterable): an iterable that yields one element of a nexus
            search response at a time, such as the one returned by
            _paginate_get().
        term (str): if defined, only items with an attribute `path` that starts
            with the given parameter are returned.
        partial_match (bool): if True, include items whose artefact path starts
            with the given term.

    Yields:
        dict: items that matched the filter.
    """
    def is_match(path_, term_):
        if partial_match:
            return path_.startswith(term_)
        else:
            return path_ == term_

    for artefact in raw_response:
        artefact_path = artefact.get('path')
        if artefact_path is None:
            continue
        if not validate_strings(artefact_path):
            continue
        if term is None or is_match(artefact_path, term):
            yield artefact


def calculate_hash(hash_name, file_path_or_handle):
    """
    Calculate a hash for the given file.

    :param hash_name: name of the hash algorithm in hashlib
    :type hash_name: str
    :param file_path_or_handle: source file name (:py:obj:`str`) or file
        handle (:py:obj:`file-like`) for the hash algorithm.
    :type file_path_or_handle: str
    :return: the calculated hash
    :rtype: str
    """
    def _hash(_fd):
        h = hashlib.new(hash_name)
        stat = os.fstat(_fd.fileno())
        if stat.st_size > 0:  # can't map a zero-length file
            m = mmap.mmap(_fd.fileno(),
                          stat.st_size, access=mmap.ACCESS_READ)
            h.update(m)
        return h.hexdigest()

    if hasattr(file_path_or_handle, 'read'):
        return _hash(file_path_or_handle)
    else:
        with open(file_path_or_handle, 'rb') as fd:
            return _hash(fd)


def has_same_hash(artefact, filepath):
    """
    Checks if a Nexus artefact has the same hash as a local filepath.

    :param artefact:  as returned by
        :py:meth:`~nexuscli.nexus_client.NexusClient.list_raw`
    :type artefact: dict
    :param filepath: local file path
    :return: True if artefact and filepath have the same hash.
    :rtype: bool
    """
    for hash_name in ['sha1', 'md5']:
        remote_hash = artefact.get('checksum', {}).get(hash_name)
        if remote_hash is None:
            continue

        local_hash = calculate_hash(hash_name, filepath)
        return local_hash == remote_hash

    return False


def ensure_exists(path, is_dir=False):
    """
    Ensures a path exists.

    :param path: the path to ensure
    :type path: pathlib.Path
    :param is_dir: whether the path is a directory.
    :type is_dir: bool
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    if is_dir:
        path.mkdir(exist_ok=True)
    else:
        path.touch()


def script_for_version(script_name, server_version, versions):
    """
    Determine if a certain nexus server version requires a different version of
    the given groovy script.

    :param script_name: original name of the script.
    :param server_version: VersionInfo for the Nexus server.
    :param versions: list of VersionInfo. Each element represents an existing
        groovy script that must be used with server_version or greater.
    :return: the version-specific name of script_name.
    """
    if server_version is None:
        return script_name

    for breaking_version in sorted(versions, reverse=True):
        if server_version >= breaking_version:
            script_path = pathlib.Path(script_name)
            # e.g.: nexus3-cli-repository-create_3.21.0.groovy
            return f'{script_path.stem}_{breaking_version}{script_path.suffix}'

    return script_name



import os
from subprocess import CalledProcessError

ENV_VAR_PREFIX = 'NEXUS3'

try:
    _, TTY_MAX_WIDTH = os.popen('stty size', 'r').read().split()
    TTY_MAX_WIDTH = int(TTY_MAX_WIDTH)
except (ValueError, CalledProcessError):
    TTY_MAX_WIDTH = 80

# see cli.util.get_client()
REQUIRED_NEXUS_OPTIONS = ['PASSWORD', 'USERNAME', 'URL']
OPTIONAL_NEXUS_OPTIONS = ['API_VERSION', 'X509_VERIFY']



import json
import semver

from nexuscli import exception, nexus_util
from nexuscli.api.repository import model

SCRIPT_CREATE_VERSIONS = [semver.VersionInfo(3, 21, 0)]
SCRIPT_NAME_CREATE = 'nexus3-cli-repository-create'
SCRIPT_NAME_DELETE = 'nexus3-cli-repository-delete'
SCRIPT_NAME_GET = 'nexus3-cli-repository-get'


def get_repository_class(raw_configuration):
    """
    Given a raw repository configuration, returns its corresponding class.

    :param raw_configuration: configuration as returned by the SCRIPT_NAME_GET
        groovy script.
    :type raw_configuration: dict
    :return: repository class
    """
    class_name = _repository_class_name(raw_configuration)
    for class_ in model.__all__:
        if class_.__name__ == class_name:
            return class_
    raise NotImplementedError(f'{class_name} for {raw_configuration}')


def _recipe_name(raw_configuration):
    """
    Given a raw repository configuration, returns its recipe name.

    :param raw_configuration: configuration as returned by the SCRIPT_NAME_GET
        groovy script.
    :type raw_configuration: dict
    :return: name of the recipe ("format")
    """
    name = raw_configuration['recipeName'].split('-')[0].title()

    if name == 'Maven2':
        name = 'Maven'

    return name


def _recipe_type(raw_configuration):
    """
    Given a raw repository configuration, returns its recipe type.

    :param raw_configuration: configuration as returned by the SCRIPT_NAME_GET
        groovy script.
    :type raw_configuration: dict
    :return: Group, Proxy or Hosted
    """
    return raw_configuration['recipeName'].split('-')[1].title()


def _repository_class_name(raw_configuration):
    """
    Given a raw repository configuration, returns the corresponding python
    repository class name.

    :param raw_configuration: configuration as returned by the SCRIPT_NAME_GET
        groovy script.
    :type raw_configuration: dict
    :return: name of the repository class
    """
    recipe_name = _recipe_name(raw_configuration)
    class_name = ''

    # The Repository class is the default because it implements most of the
    # recipes
    if recipe_name not in model.Repository.RECIPES:
        class_name += recipe_name

    class_name += _recipe_type(raw_configuration)

    class_name += 'Repository'

    return class_name


# TODO: flattening-out the configuration on the python class turned-out to be
#   a bad idea because now we need to convert it back and forth. Perhaps add
#   a compatible change that allows one to provide the flat kwargs OR a
#   configuration dict in the format accepted/returned by Nexus.
def _add_proxy_kwargs(kwargs, attributes):
    kwargs['auto_block'] = attributes['httpclient']['autoBlock']

    kwargs['content_max_age'] = int(attributes['proxy']['contentMaxAge'])
    kwargs['metadata_max_age'] = int(attributes['proxy']['metadataMaxAge'])
    kwargs['remote_url'] = attributes['proxy']['remoteUrl']

    kwargs['negative_cache_enabled'] = attributes['negativeCache']['enabled']
    kwargs['negative_cache_ttl'] = int(
        attributes['negativeCache']['timeToLive'])


def _add_maven_kwargs(kwargs, attributes):
    kwargs['layout_policy'] = attributes['maven']['layoutPolicy']
    kwargs['version_policy'] = attributes['maven']['versionPolicy']


def _add_yum_kwargs(kwargs, attributes):
    kwargs['depth'] = int(attributes['yum']['repodataDepth'])
    # TODO: support yum deploy policy
    # kwargs['TODO'] = attributes['yum']['deployPolicy']


def _add_apt_kwargs(kwargs, attributes):
    if 'aptSigning' in attributes:
        kwargs['passphrase'] = attributes['aptSigning']['passphrase']
        kwargs['gpg_keypair'] = attributes['aptSigning']['keypair']
    if 'apt' in attributes:
        kwargs['distribution'] = attributes['apt']['distribution']
        if attributes['apt'].get('flat') is not None:
            kwargs['flat'] = attributes['apt']['flat']


def _add_hosted_kwargs(kwargs, attributes):
    kwargs['write_policy'] = attributes['storage']['writePolicy']


def _add_common_kwargs(kwargs, attributes):
    kwargs['cleanup_policy'] = attributes.get('cleanup', {}).get('policyName')
    if kwargs['cleanup_policy'] == 'None':
        kwargs['cleanup_policy'] = None

    kwargs['blob_store_name'] = attributes['storage']['blobStoreName']
    kwargs['strict_content_type_validation'] = attributes[
        'storage']['strictContentTypeValidation']


def _repository_args_kwargs(raw_configuration):
    args = (raw_configuration['repositoryName'],)
    kwargs = {'recipe': _recipe_name(raw_configuration)}

    attributes = raw_configuration['attributes']
    _add_common_kwargs(kwargs, attributes)

    if _recipe_name(raw_configuration) == 'Maven':
        _add_maven_kwargs(kwargs, attributes)

    if _recipe_name(raw_configuration) == 'Yum':
        _add_yum_kwargs(kwargs, attributes)

    if _recipe_name(raw_configuration) == 'Apt':
        _add_apt_kwargs(kwargs, attributes)

    # TODO: support Group
    if _recipe_type(raw_configuration) == 'Proxy':
        _add_proxy_kwargs(kwargs, attributes)

    if _recipe_type(raw_configuration) == 'Hosted':
        _add_hosted_kwargs(kwargs, attributes)

    return args, kwargs


class RepositoryCollection:
    """
    A class to manage Nexus 3 repositories.

    Args:
        client(nexuscli.nexus_client.NexusClient): the client instance that
            will be used to perform operations against the Nexus 3 service. You
            must provide this at instantiation or set it before calling any
            methods that require connectivity to Nexus.
    """

    def __init__(self, client=None):
        self._client = client
        self._repositories_json = None

    def get_by_name(self, name):
        """
        Get a Nexus 3 repository by its name.

        :param name: name of the repository wanted
        :type name: str
        :rtype: nexuscli.api.repository.model.Repository
        :raise exception.NexusClientInvalidRepository: when a repository with
            the given name isn't found.
        """
        configuration = self.get_raw_by_name(name)
        Repository = get_repository_class(configuration)
        args, kwargs = _repository_args_kwargs(configuration)

        return Repository(*args, nexus_client=self._client, **kwargs)

    def get_raw_by_name(self, name):
        """
        Return the raw dict for the repository called ``name``. Remember to
        :meth:`refresh` to get the latest from the server.

        :param name: name of the repository wanted
        :type name: str
        :rtype: dict
        :raise exception.NexusClientInvalidRepository: when a repository with
            the given name isn't found.
        """
        self._client.scripts.create_if_missing(SCRIPT_NAME_GET)

        resp = self._client.scripts.run(SCRIPT_NAME_GET, data=name)
        configuration = resp.get('result')

        if configuration == 'null':
            raise exception.NexusClientInvalidRepository(name)

        return json.loads(configuration)

    def refresh(self):
        """
        Refresh local list of repositories with latest from service. A raw
        representation of repositories can be fetched using :meth:`raw_list`.
        """
        response = self._client.http_get('repositories')
        if response.status_code != 200:
            raise exception.NexusClientAPIError(response.content)

        self._repositories_json = response.json()

    def raw_list(self):
        """
        A raw representation of the Nexus repositories.

        Returns:
            dict: for the format, see `List Repositories
            <https://help.sonatype.com/repomanager3/rest-and-integration-api/repositories-api#RepositoriesAPI-ListRepositories>`_.
        """
        self.refresh()
        return self._repositories_json

    def delete(self, name):
        """
        Delete a repository.

        :param name: name of the repository to be deleted.
        :type name: str
        """
        self._client.scripts.create_if_missing(SCRIPT_NAME_DELETE)
        self._client.scripts.run(SCRIPT_NAME_DELETE, data=name)

    def create(self, repository):
        """
        Creates a Nexus repository with the given format and type.

        :param repository: the instance containing the settings for the
            repository to be created.
        :type repository: Repository
        :raises NexusClientCreateRepositoryError: error creating repository.
        """
        script_name = nexus_util.script_for_version(
            SCRIPT_NAME_CREATE,
            self._client.server_version,
            SCRIPT_CREATE_VERSIONS)

        if not issubclass(type(repository), model.Repository):
            raise TypeError(f'{repository} has type {type(repository)}'
                            f' but must be a subclass of Repository')

        self._client.scripts.create_if_missing(script_name)

        script_args = json.dumps(repository.configuration)
        resp = self._client.scripts.run(script_name, data=script_args)

        result = resp.get('result')
        if result != 'null':
            raise exception.NexusClientCreateRepositoryError(resp)



"""Methods to implement upload for specific repository formats (recipes)"""
import os

from nexuscli import exception
from nexuscli.api.repository.validations import REMOTE_PATH_SEPARATOR


def upload_file_raw(repository, src_file, dst_dir, dst_file):
    """
    Upload a single file to a raw repository.

    :param repository: repository instance used to access Nexus 3 service.
    :type repository: nexuscli.api.repository.model.Repository
    :param src_file: path to the local file to be uploaded.
    :param dst_dir: directory under dst_repo to place file in. When None,
        the file is placed under the root of the raw repository
    :param dst_file: destination file name.
    :raises exception.NexusClientInvalidRepositoryPath: invalid repository
        path.
    :raises exception.NexusClientAPIError: unknown response from Nexus API.
    """
    dst_dir = os.path.normpath(dst_dir or REMOTE_PATH_SEPARATOR)

    params = {'repository': repository.name}
    files = {'raw.asset1': open(src_file, 'rb').read()}
    data = {
        'raw.directory': dst_dir,
        'raw.asset1.filename': dst_file,
    }

    response = repository.nexus_client.http_post(
        'components', files=files, data=data, params=params, stream=True)

    if response.status_code != 204:
        raise exception.NexusClientAPIError(
            f'Uploading to {repository.name}. Reason: {response.reason} '
            f'Status code: {response.status_code} Text: {response.text}')


def upload_file_yum(repository, src_file, dst_dir, dst_file):
    """
    Upload a single file to a yum repository.

    :param repository: repository instance used to access Nexus 3 service.
    :type repository: nexuscli.api.repository.model.Repository
    :param src_file: path to the local file to be uploaded.
    :param dst_dir: directory under dst_repo to place file in.
    :param dst_file: destination file name.
    :raises exception.NexusClientAPIError: unknown response from Nexus API.
    """
    dst_dir = dst_dir or REMOTE_PATH_SEPARATOR
    repository_path = REMOTE_PATH_SEPARATOR.join(
        ['repository', repository.name, dst_dir, dst_file])

    with open(src_file, 'rb') as fh:
        response = repository.nexus_client.http_put(
            repository_path, data=fh, stream=True,
            service_url=repository.nexus_client.config.url)

    if response.status_code != 200:
        raise exception.NexusClientAPIError(
            f'Uploading to {repository_path}. Reason: {response.reason} '
            f'Status code: {response.status_code} Text: {response.text}')



"""Handles base/root commands (as opposed to subcommands)"""
import inflect
import sys
import types

from nexuscli import exception, nexus_config
from nexuscli.nexus_client import NexusClient


PLURAL = inflect.engine().plural


def cmd_login(**kwargs):
    """Performs ``nexus3 login``"""
    config = nexus_config.NexusConfig(**kwargs)

    # make sure configuration works before saving
    try:
        NexusClient(config=config)
    except exception.NexusClientInvalidCredentials:
        # the regular message tells the user to try to login, which is what
        # they just did now, so override the msg
        raise exception.NexusClientInvalidCredentials('Invalid credentials')

    sys.stderr.write(f'\nLogin successful.\n')
    sys.stderr.write(f'Configuration saved to {config.dump()}, {config.dump_env()}\n')

    return exception.CliReturnCode.SUCCESS.value


def cmd_list(nexus_client, repository_path):
    """Performs ``nexus3 list``"""
    artefact_list = nexus_client.list(repository_path)

    # FIXME: is types.GeneratorType still used?
    if isinstance(artefact_list, (list, types.GeneratorType)):
        for artefact in iter(artefact_list):
            print(artefact)
        return exception.CliReturnCode.SUCCESS.value
    else:
        return exception.CliReturnCode.UNKNOWN_ERROR.value


def _cmd_up_down_errors(count, action):
    """Print and exit with error if upload/download/delete didn't succeed"""
    if count == 0:
        # FIXME: inflex the action verb to past participle
        sys.stderr.write('WARNING: no files were {}\'ed.'.format(action))
        sys.exit(exception.CliReturnCode.NO_FILES.value)

    if count == -1:
        sys.stderr.write('ERROR during {} operation.'.format(action))
        sys.exit(exception.CliReturnCode.API_ERROR.value)


def cmd_upload(nexus_client, src=None, dst=None, flatten=None, recurse=None):
    """Performs ``nexus3 upload``"""
    sys.stderr.write(f'Uploading {src} to {dst}\n')

    upload_count = nexus_client.upload(
        src, dst, flatten=flatten, recurse=recurse)

    _cmd_up_down_errors(upload_count, 'upload')

    file = PLURAL('file', upload_count)
    sys.stderr.write(f'Uploaded {upload_count} {file} to {dst}\n')
    return exception.CliReturnCode.SUCCESS.value


def cmd_download(nexus_client, src=None, dst=None, flatten=None, cache=None):
    """Performs ``nexus3 download``"""
    sys.stderr.write(f'Downloading {src} to {dst}\n')

    download_count = nexus_client.download(
        src, dst, flatten=flatten, nocache=not cache)

    _cmd_up_down_errors(download_count, 'download')

    file_word = PLURAL('file', download_count)
    sys.stderr.write(
        f'Downloaded {download_count} {file_word} to {dst}\n')
    return exception.CliReturnCode.SUCCESS.value


def cmd_delete(nexus_client, repository_path):
    """Performs ``nexus3 delete``"""
    delete_count = nexus_client.delete(repository_path)

    _cmd_up_down_errors(delete_count, 'delete')

    file_word = PLURAL('file', delete_count)
    sys.stderr.write(f'Deleted {delete_count} {file_word}\n')
    return exception.CliReturnCode.SUCCESS.value



from click import ClickException
from enum import Enum


class CliReturnCode(Enum):
    """Error codes returned by :py:mod:`nexuscli.cli`"""
    SUCCESS = 0
    NO_FILES = 1
    API_ERROR = 2
    CONNECTION_ERROR = 3
    DOWNLOAD_ERROR = 4
    INVALID_CREDENTIALS = 5
    INVALID_SUBCOMMAND = 10
    SUBCOMMAND_ERROR = 11
    POLICY_NOT_FOUND = 20
    REPOSITORY_NOT_FOUND = 30
    UNKNOWN_ERROR = 99


class NexusClientBaseError(ClickException):
    exit_code = CliReturnCode.UNKNOWN_ERROR.value

    # TODO: add error message; e.g.:
    # nexus_error = CliReturnCode.UNKNOWN_ERROR
    # def __init__(self, message):
    #     super().__init__(self.nexus_error.message)


class NexusClientAPIError(NexusClientBaseError):
    """Unexpected response from Nexus service."""
    exit_code = CliReturnCode.API_ERROR.value


class NexusClientConnectionError(NexusClientBaseError):
    """Generic network connector error."""
    exit_code = CliReturnCode.CONNECTION_ERROR.value


class NexusClientInvalidCredentials(NexusClientBaseError):
    """
    Login credentials not accepted by Nexus service. Usually the result of a
    HTTP 401 response.
    """
    exit_code = CliReturnCode.INVALID_CREDENTIALS.value


class NexusClientInvalidRepositoryPath(NexusClientBaseError):
    """
    Used when an operation against the Nexus service uses an invalid or
    non-existent path.
    """
    pass


class NexusClientInvalidRepository(NexusClientBaseError):
    """The given repository does not exist in Nexus."""
    exit_code = CliReturnCode.REPOSITORY_NOT_FOUND.value


class NexusClientInvalidCleanupPolicy(NexusClientBaseError):
    """The given cleanup policy does not exist in Nexus."""
    exit_code = CliReturnCode.SUBCOMMAND_ERROR.value


class NexusClientCreateRepositoryError(NexusClientBaseError):
    """Used when a repository creation operation in Nexus fails."""
    exit_code = CliReturnCode.SUBCOMMAND_ERROR.value


class NexusClientCreateCleanupPolicyError(NexusClientBaseError):
    """Used when a cleanup policy creation operation in Nexus fails."""
    exit_code = CliReturnCode.SUBCOMMAND_ERROR.value


class DownloadError(NexusClientBaseError):
    """Error retrieving artefact from Nexus service."""
    exit_code = CliReturnCode.DOWNLOAD_ERROR.value



from texttable import Texttable

from nexuscli import exception
from nexuscli.cli import constants


def cmd_list(nexus_client):
    """Performs ``nexus3 script list``"""
    scripts = nexus_client.scripts.list()

    table = Texttable(max_width=constants.TTY_MAX_WIDTH)
    table.add_row(['Name', 'Type', 'Content'])
    table.set_deco(Texttable.HEADER | Texttable.HLINES)
    for script in scripts:
        content = script['content']
        if len(content) > 40:
            content = f'{content[:40]}...'
        table.add_row([script['name'], script['type'], content])

    print(table.draw())
    return exception.CliReturnCode.SUCCESS.value


def cmd_create(nexus_client, name, content, **kwargs):
    """Performs ``nexus3 script create``"""
    nexus_client.scripts.create(name, content, **kwargs)
    return exception.CliReturnCode.SUCCESS.value


def cmd_delete(nexus_client, name):
    """Performs ``nexus3 script delete``"""
    nexus_client.scripts.delete(name)
    return exception.CliReturnCode.SUCCESS.value


def cmd_run(nexus_client, name, arguments):
    """Performs ``nexus3 script run``"""
    resp = nexus_client.scripts.run(name, arguments)
    print(resp)
    return exception.CliReturnCode.SUCCESS.value



class CleanupPolicy(object):
    """
    Represents a Nexus Cleanup Policy.

    Example structure and attributes common to all repositories:

        >>> kwargs = {
        >>>     'name': 'my-policy',
        >>>     'format': 'bower',
        >>>     'notes': 'Some comment',
        >>>     'criteria': {
        >>>         'lastDownloaded': 172800,
        >>>         'lastBlobUpdated': 86400,
        >>>         'regex': 'matchthis'
        >>>     }
        >>> }


    Args:
        client (nexuscli.nexus_client.NexusClient): the client instance that
            will be used to perform operations against the Nexus 3 service. You
            must provide this at instantiation or set it before calling any
            methods that require connectivity to Nexus.
        name (str): name of the new policy.
        format (str): 'all' or the format of the repository this policy
            applies to.
        lastDownloaded (int): deletion criterion: days since artefact last
            downloaded
        lastBlobUpdated (int): deletion criterion: days since last update to
            artefact
        regex (str): deletion criterion: only delete artefacts that match this
            regular expression
    """
    def __init__(self, client, **kwargs):
        self._client = client
        # TODO: validate kwargs
        self._raw = kwargs

    @property
    def configuration(self):
        """
        Nexus 3 Cleanup Policy representation as a python dict. The dict
        returned by this property can
        be converted to JSON for use with the ``nexus3-cli-cleanup-policy``
        groovy script created by the
        :py:class:`~nexuscli.api.cleanup_policy.collection.CleanupPolicyCollection`
        methods.

        :return: cleanup policy as a dict
        :rtype: dict
        """
        return self._raw



import click
import logging
import pkg_resources

from nexuscli import (
    LOG_LEVEL,
    nexus_config)
from nexuscli.api.repository import model
from nexuscli.cli import (
    root_commands, util, subcommand_repository,
    subcommand_cleanup_policy, subcommand_script)
from nexuscli.cli.constants import ENV_VAR_PREFIX

PACKAGE_VERSION = pkg_resources.get_distribution('nexus3-cli').version
CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help'], auto_envvar_prefix=ENV_VAR_PREFIX)

logging.basicConfig(level=LOG_LEVEL)


#############################################################################
# root commands
@click.group(cls=util.AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(version=PACKAGE_VERSION, message='%(version)s')
def nexus_cli():
    pass


@nexus_cli.command()
@click.option(
    '--url', '-U', default=nexus_config.DEFAULTS['url'], prompt=True,
    help='Nexus OSS URL', show_default=True, allow_from_autoenv=True, show_envvar=True)
@click.option(
    '--username', '-u', default=nexus_config.DEFAULTS['username'], prompt=True,
    help='Nexus user', show_default=True, allow_from_autoenv=True, show_envvar=True)
@click.option(
    '--password', '-p', prompt=True, hide_input=True,
    help='Password for user', allow_from_autoenv=True, show_envvar=True)
@click.option(
    '--x509_verify/--no-x509_verify', prompt=True,
    default=nexus_config.DEFAULTS['x509_verify'], show_default=True,
    help='Verify server certificate', allow_from_autoenv=True, show_envvar=True)
def login(**kwargs):
    """
    Login to Nexus server, saving settings to ~/.nexus-cli.
    """
    root_commands.cmd_login(**kwargs)


@nexus_cli.command(name='list')
@click.argument('repository_path')
@util.with_nexus_client
def list_(ctx: click.Context, repository_path):
    """
    List all files within REPOSITORY_PATH.

    REPOSITORY_PATH must start with a repository name.
    """
    root_commands.cmd_list(ctx.obj, repository_path)


@nexus_cli.command()
@click.argument('repository_path')
@util.with_nexus_client
def delete(ctx: click.Context, repository_path):
    """
    Recursively delete all files under REPOSITORY_PATH.

    REPOSITORY_PATH must start with a repository name.
    """
    root_commands.cmd_delete(ctx.obj, repository_path)


@nexus_cli.command()
# TODO: use Path for src argument
@click.argument('src')
@click.argument('dst')
@click.option('--flatten/--no-flatten', default=False,
              help='Flatten DST directory structure')
@click.option('--recurse/--no-recurse', default=True,
              help='Process all SRC subdirectories')
@util.with_nexus_client
def upload(ctx: click.Context, **kwargs):
    """
    Upload local SRC to remote DST.  If either argument ends with a `/`, it's
    assumed to be a directory.

    DEST must start with a repository name and optionally be followed by the
    path where SRC is to be uploaded to.
    """
    root_commands.cmd_upload(ctx.obj, **kwargs)


@nexus_cli.command()
@click.argument('src')
@click.argument('dst')
@click.option('--flatten/--no-flatten', default=False,
              help='Flatten DEST directory structure')
@click.option('--cache/--no-cache', default=True,
              help='Do not download if a local copy is already up-to-date')
@util.with_nexus_client
def download(ctx: click.Context, **kwargs):
    """
    Download remote SRC to local DEST.  If either argument ends with a `/`,
    it's assumed to be a directory.

    SRC must start with a repository name and optionally be followed by a path
    to be downloaded.
    """
    root_commands.cmd_download(ctx.obj, **kwargs)


#############################################################################
# repository sub-commands
@nexus_cli.group(cls=util.AliasedGroup)
def repository():
    """
    Manage repositories.
    """
    pass


@repository.command(name='list')
@util.with_nexus_client
def repository_list(ctx: click.Context):
    """
    List all repositories.
    """
    subcommand_repository.cmd_list(ctx.obj)


@repository.command(name='show')
@click.argument('repository_name')
@util.with_nexus_client
def repository_show(ctx: click.Context, repository_name):
    """
    Show the configuration for REPOSITORY_NAME as JSON.
    """
    subcommand_repository.cmd_show(ctx.obj, repository_name)


@repository.command(name='delete')
@click.argument('repository_name')
@click.confirmation_option()
@util.with_nexus_client
def repository_delete(ctx: click.Context, repository_name):
    """
    Delete REPOSITORY_NAME (but not its blobstore).
    """
    subcommand_repository.cmd_delete(ctx.obj, repository_name)


#############################################################################
# repository create sub-commands
@repository.group(cls=util.AliasedGroup, name='create')
def repository_create():
    """
    Create a repository.
    """
    pass


#############################################################################
# repository create hosted sub-commands
@repository_create.command(
    name='hosted',
    cls=util.mapped_commands({
        'apt': model.AptHostedRepository.RECIPES,
        'docker': model.DockerHostedRepository.RECIPES,
        'maven': model.MavenHostedRepository.RECIPES,
        'yum': model.YumHostedRepository.RECIPES,
        # generic, remaining repositories
        'recipe': model.HostedRepository.RECIPES,
    }))
def repository_create_hosted():
    """
    Create a hosted repository.
    """
    pass


def _create_repository(ctx, repo_type, **kwargs):
    # every repository recipe needs these
    kwargs['recipe'] = ctx.info_name
    util.upcase_values(
        kwargs, ['index_type', 'layout_policy', 'version_policy',
                 'write_policy'])

    # these CLI options were shortened for user convenience; fix them now
    util.rename_keys(kwargs, {
        'negative_cache': 'negative_cache_enabled',
        'strict_content': 'strict_content_type_validation',
        'trust_store': 'use_trust_store_for_index_access',
    })

    subcommand_repository.cmd_create(ctx, repo_type=repo_type, **kwargs)


REPOSITORY_COMMON_OPTIONS = [
    click.argument('repository-name'),
    click.option('--blob-store-name', default='default',
                 help='Blobstore name to use with new repository'),
    click.option('--strict-content/--no-strict-content', default=False,
                 help='Toggle strict content type validation'),
    click.option('--cleanup-policy',
                 help='Name of existing clean-up policy to use'),
]

REPOSITORY_COMMON_HOSTED_OPTIONS = REPOSITORY_COMMON_OPTIONS + [
    click.option(
        '--write-policy', help='Write policy to use', default='allow',
        type=click.Choice(['allow', 'allow_once', 'deny'],
                          case_sensitive=False))
]


@repository_create_hosted.command(name='recipe')
@util.add_options(REPOSITORY_COMMON_HOSTED_OPTIONS)
@util.with_nexus_client
def repository_create_hosted_recipe(ctx: click.Context, **kwargs):
    """
    Create a hosted repository.
    """
    _create_repository(ctx, 'hosted', **kwargs)


REPOSITORY_COMMON_APT_OPTIONS = [
    click.option(
        '--distribution', required=True,
        help='Distribution to fetch; e.g.: bionic')
]


@repository_create_hosted.command(name='apt')
@util.add_options(REPOSITORY_COMMON_HOSTED_OPTIONS)
@util.add_options(REPOSITORY_COMMON_APT_OPTIONS)
@click.option(
    '--gpg-keypair', required=True, type=click.File(),
    default='./private.gpg.key', help='Path to GPG signing key')
@click.option('--passphrase', help='Passphrase for GPG key pair')
@util.with_nexus_client
def repository_create_hosted_apt(ctx: click.Context, **kwargs):
    """
    Create a hosted apt repository.
    """
    kwargs['gpg_keypair'] = kwargs['gpg_keypair'].read()
    _create_repository(ctx, 'hosted', **kwargs)


REPOSITORY_COMMON_DOCKER_OPTIONS = [
    click.option(
        '--v1-enabled/--no-v1-enabled', default=False,
        help='Enable v1 registry'),
    click.option(
        '--force-basic-auth/--no-force-basic-auth', default=False,
        help='Force use of basic authentication'),
    click.option(
        '--http-port', type=click.INT, help='Port for HTTP service'),
    click.option(
        '--https-port', type=click.INT, help='Port for HTTPS service'),
]


@repository_create_hosted.command(name='docker')
@util.add_options(REPOSITORY_COMMON_HOSTED_OPTIONS)
@util.add_options(REPOSITORY_COMMON_DOCKER_OPTIONS)
@util.with_nexus_client
def repository_create_hosted_docker(ctx: click.Context, **kwargs):
    """
    Create a hosted docker repository.
    """
    _create_repository(ctx, 'hosted', **kwargs)


REPOSITORY_COMMON_MAVEN_OPTIONS = [
    click.option(
        '--version-policy', help='Version policy to use', default='release',
        type=click.Choice(['release', 'snapshot', 'mixed'],
                          case_sensitive=False)),
    click.option(
        '--layout-policy', help='Layout policy to use', default='strict',
        type=click.Choice(['strict', 'permissive'],
                          case_sensitive=False)),
]


@repository_create_hosted.command(name='maven')
@util.add_options(REPOSITORY_COMMON_HOSTED_OPTIONS)
@util.add_options(REPOSITORY_COMMON_MAVEN_OPTIONS)
@util.with_nexus_client
def repository_create_hosted_maven(ctx: click.Context, **kwargs):
    """
    Create a hosted maven repository.
    """
    _create_repository(ctx, 'hosted', **kwargs)


@repository_create_hosted.command(name='yum')
@util.add_options(REPOSITORY_COMMON_HOSTED_OPTIONS)
@click.option(
    '--depth', help='Depth where repodata folder(s) exist', default=0,
    type=click.IntRange(min=0, max=5, clamp=False))
@util.with_nexus_client
def repository_create_hosted_yum(ctx: click.Context, **kwargs):
    """
    Create a hosted yum repository.
    """
    _create_repository(ctx, 'hosted', **kwargs)


#############################################################################
# repository create proxy sub-commands
@repository_create.command(
    name='proxy',
    cls=util.mapped_commands({
        'apt': model.AptProxyRepository.RECIPES,
        'docker': model.DockerProxyRepository.RECIPES,
        'maven': model.MavenProxyRepository.RECIPES,
        'yum': model.YumProxyRepository.RECIPES,
        # remaining, generic repositories
        'recipe': model.ProxyRepository.RECIPES,
    }))
def repository_create_proxy():
    """
    Create a proxy repository.
    """
    pass


REPOSITORY_COMMON_PROXY_OPTIONS = REPOSITORY_COMMON_OPTIONS + [
    click.argument('remote-url'),
    click.option(
        '--auto-block/--no-auto-block', default=True,
        help='Disable outbound connections on remote-url access errors'),
    click.option(
        '--negative-cache/--no-negative-cache', default=True,
        help='Cache responses for content missing in the remote-url'),
    click.option(
        '--negative-cache-ttl', type=click.INT, default=1440,
        help='Cache time in minutes'),
    click.option(
        '--content-max-age', type=click.INT, default=1440,
        help='Maximum age of cached artefacts'),
    click.option(
        '--metadata-max-age', type=click.INT, default=1440,
        help='Maximum age of cached artefacts metadata'),
    click.option(
        '--remote-auth-type', help='Only username is supported',
        type=click.Choice(['username'], case_sensitive=False)),
    # TODO: require `--remote-auth-type username` when these are specified
    click.option('--remote-username', help='Username for remote URL'),
    click.option('--remote-password', help='Password for remote URL'),
]


@repository_create_proxy.command(name='recipe')
@util.add_options(REPOSITORY_COMMON_PROXY_OPTIONS)
@util.with_nexus_client
def repository_create_proxy_recipe(ctx: click.Context, **kwargs):
    """
    Create a proxy repository.
    """
    _create_repository(ctx, 'proxy', **kwargs)


@repository_create_proxy.command(name='apt')
@util.add_options(REPOSITORY_COMMON_PROXY_OPTIONS)
@util.add_options(REPOSITORY_COMMON_APT_OPTIONS)
@click.option(
    '--flat/--no-flat', default=False, help='Is this repository flat?')
@util.with_nexus_client
def repository_create_proxy_apt(ctx: click.Context, **kwargs):
    """
    Create a proxy apt repository.
    """
    _create_repository(ctx, 'proxy', **kwargs)


@repository_create_proxy.command(name='docker')
@util.add_options(REPOSITORY_COMMON_PROXY_OPTIONS)
@util.add_options(REPOSITORY_COMMON_DOCKER_OPTIONS)
@click.option(
    '--index-type', help='Docker index type', default='registry',
    type=click.Choice(['registry', 'hub', 'custom'],
                      case_sensitive=False))
# TODO: enforce requirement
@click.option('--index-url', help='Required for --index-type custom')
@click.option(
    '--trust-store/--no-trust-store', default=False,
    help='Required for --index-type hub or custom')
@util.with_nexus_client
def repository_create_proxy_docker(ctx: click.Context, **kwargs):
    """
    Create a docker proxy repository.
    """
    _create_repository(ctx, 'proxy', **kwargs)


@repository_create_proxy.command(name='maven')
@util.add_options(REPOSITORY_COMMON_PROXY_OPTIONS)
@util.add_options(REPOSITORY_COMMON_MAVEN_OPTIONS)
@util.with_nexus_client
def repository_create_proxy_maven(ctx: click.Context, **kwargs):
    """
    Create a maven proxy repository.
    """
    _create_repository(ctx, 'proxy', **kwargs)


@repository_create_proxy.command(name='yum')
@util.add_options(REPOSITORY_COMMON_PROXY_OPTIONS)
@util.with_nexus_client
def repository_create_proxy_yum(ctx: click.Context, **kwargs):
    """
    Create a yum proxy repository.
    """
    _create_repository(ctx, 'proxy', **kwargs)


#############################################################################
# cleanup_policy sub-commands
@nexus_cli.group(cls=util.AliasedGroup)
def cleanup_policy():
    """
    Manage clean-up policies.
    """
    pass


@cleanup_policy.command(name='create')
@click.argument('name')
@click.option(
    '--format', default='all',
    help='The recipe that this cleanup policy can be applied to',
    type=click.Choice(['all'] + list(model.SUPPORTED_FORMATS)))
@click.option(
    '--downloaded', type=click.IntRange(min=1),
    help='Cleanup criteria; last downloaded in this many days.')
@click.option(
    '--updated', type=click.IntRange(min=1),
    help='Cleanup criteria; last updated in this many days.')
# TODO: validate formats that accept regex
@click.option(
    '--regex',
    help='Cleanup criteria; only cleanup components matching expression')
@click.option(
    '--notes',
    help='Notes about your policy')
@util.with_nexus_client
def cleanup_policy_create(ctx: click.Context, **kwargs):
    """
    Create or update a cleanup policy called NAME.
    """
    # TODO: use a click type for this check?
    criteria_keys = {'downloaded', 'updated', 'regex'}
    util.move_to_key(kwargs, 'criteria', criteria_keys)

    util.rename_keys(kwargs['criteria'], {
        'downloaded': 'lastDownloaded',
        'updated': 'lastBlobUpdated',
    })

    subcommand_cleanup_policy.cmd_create(ctx.obj, **kwargs)


@cleanup_policy.command(name='list')
@util.with_nexus_client
def cleanup_policy_list(ctx: click.Context):
    subcommand_cleanup_policy.cmd_list(ctx.obj)


#############################################################################
# script sub-commands
@nexus_cli.group(cls=util.AliasedGroup)
def script():
    """
    Manage scripts.
    """
    pass


@script.command(name='create')
@click.argument('name')
@click.argument('file', type=click.File())
@click.option('--script-type', default='groovy', help='Script type')
@util.with_nexus_client
def script_create(ctx: click.Context, name, file, **kwargs):
    """
    Create a new script called NAME from FILE.
    """
    subcommand_script.cmd_create(ctx.obj, name, file.read(), **kwargs)


@script.command(name='delete')
@click.argument('name')
@util.with_nexus_client
def script_delete(ctx: click.Context, name):
    """
    Delete the script called NAME.
    """
    subcommand_script.cmd_delete(ctx.obj, name)


@script.command(name='list')
@util.with_nexus_client
def script_list(ctx: click.Context):
    """
    List all scripts.
    """
    subcommand_script.cmd_list(ctx.obj)


@script.command(name='run')
@click.argument('name')
@click.option('--script-arguments', '-a')
@util.with_nexus_client
def script_run(ctx: click.Context, name, script_arguments):
    """
    Run the script called NAME.
    """
    subcommand_script.cmd_run(ctx.obj, name, script_arguments)



import os

from nexuscli.api.repository.validations import REMOTE_PATH_SEPARATOR


def get_files(src_dir, recurse=True):
    """
    Walks the given directory and collects files to be uploaded. If
    recurse option is False, only the files on the root of the directory
    will be returned.

    :param src_dir: location of files
    :param recurse: If false, only the files on the root of src_dir
                    are returned
    :return: file set to be used with upload_directory
    :rtype: set
    """
    source_files = set()
    for dirname, dirnames, filenames in os.walk(src_dir):
        if not recurse:
            del dirnames[:]

        source_files.update(
            os.path.relpath(os.path.join(dirname, f), src_dir)
            for f in filenames)

    return source_files


def get_upload_subdirectory(dst_dir, file_path, flatten=False):
    """
    Find the destination subdirectory based on given parameters. This is mostly
    so the `flatten` option is honoured.

    :param dst_dir: destination directory
    :param file_path: file path, using REMOTE_PATH_SEPARATOR as the directory
        separator.
    :param flatten: when True, sub_directory will be flattened (ie: file_path
        structure will not be present in the destination directory)
    :type flatten: bool
    :return: the appropriate sub directory in the destination directory.
    :rtype: str
    """
    # empty dst_dir because most repo formats, aside from raw, allow it
    sub_directory = dst_dir or ''
    if flatten:
        return sub_directory

    sep = REMOTE_PATH_SEPARATOR
    dirname = os.path.dirname(file_path)
    if sub_directory.endswith(sep) or dirname.startswith(sep):
        sep = ''
    sub_directory += f'{sep}{dirname}'

    return sub_directory



import click
import functools
import os
import sys
import typing

from nexuscli import exception
from nexuscli.cli import constants
from nexuscli.nexus_client import NexusClient
from nexuscli.nexus_config import NexusConfig


class AliasedGroup(click.Group):
    """
    Implements execution of the first partial match for a command. Fails with a
    message if there are no unique matches.

    See: https://click.palletsprojects.com/en/7.x/advanced/#command-aliases
    """
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


def with_nexus_client(click_command):
    @functools.wraps(click_command)
    @click.pass_context
    def command(ctx: click.Context, **kwargs):
        ctx.obj = get_client()
        return click_command(ctx, **kwargs)

    return command


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


def move_to_key(mydict, new_key, keys_to_move):
    if set(mydict.keys()).intersection(keys_to_move):
        mydict[new_key] = {}
        for k in keys_to_move:
            mydict[new_key][k] = mydict[k]
            del mydict[k]


def mapped_commands(command_map: dict):
    """
    TODO: document command_map format

    :param command_map:
    :return:
    """
    class CommandGroup(click.Group):
        def get_command(self, ctx, cmd_name):
            for real_command, aliases in command_map.items():
                if cmd_name in aliases:
                    return click.Group.get_command(self, ctx, real_command)
            return None

        def list_commands(self, ctx):
            return [a for b in command_map.values() for a in b]

    return CommandGroup


def upcase_values(mydict: dict, keys=[]):
    for key in keys:
        value = mydict.get(key)
        if value is not None:
            mydict[key] = value.upper()


def rename_keys(mydict: dict, rename_map: dict):
    for current_name, new_name in rename_map.items():
        if mydict.get(current_name) is not None:
            mydict[new_name] = mydict[current_name]
            del mydict[current_name]


def _get_client_kwargs() -> typing.Optional[typing.Dict[str, str]]:
    def _without_prefix(name):
        return name[len(constants.ENV_VAR_PREFIX) + 1:].lower()

    def _with_prefix(names):
        return [f'{constants.ENV_VAR_PREFIX}_{x}' for x in names]

    # This seemed an easier implementation compared to "exposing" the cli.login options to all the
    # other commands. The part I don't love is that I have to repeat the option names here.
    required_variables = _with_prefix(constants.REQUIRED_NEXUS_OPTIONS)
    defined_env_vars = [x in os.environ for x in required_variables]

    if any(defined_env_vars):
        if not all(defined_env_vars):
            errmsg = 'If any of these environment variables are set, then ALL must be set: ' \
                     f'{required_variables}'
            raise exception.NexusClientInvalidCredentials(errmsg)
        else:
            config_kwargs = {}
            all_env_vars = required_variables + _with_prefix(constants.OPTIONAL_NEXUS_OPTIONS)
            for env_var in all_env_vars:
                if os.environ.get(env_var):
                    config_kwargs[_without_prefix(env_var)] = os.environ[env_var]

            return config_kwargs


def get_client():
    """
    Returns a Nexus Client instance. Prints a warning if the configuration file doesn't exist.

    :rtype: nexuscli.nexus_client.NexusClient
    """
    maybe_config = _get_client_kwargs()
    if maybe_config:
        config = NexusConfig(**_get_client_kwargs())
        return NexusClient(config=config)

    config = NexusConfig()
    try:
        config.load()
    except FileNotFoundError:
        sys.stderr.write(
            'Warning: configuration not found; proceeding with defaults.\n'
            'To remove this warning, please run `nexus3 login`\n')
    return NexusClient(config=config)



import json
import logging
import semver

from nexuscli import exception, nexus_util
from nexuscli.api.cleanup_policy import CleanupPolicy

LOG = logging.getLogger(__name__)
GROOVY_SCRIPT_VERSIONS = [semver.VersionInfo(3, 20, 0)]


class CleanupPolicyCollection(object):
    """
    A class to manage Nexus 3 Cleanup Policies.

    Args:
        client(nexuscli.nexus_client.NexusClient): the client instance that
            will be used to perform operations against the Nexus 3 service. You
            must provide this at instantiation or set it before calling any
            methods that require connectivity to Nexus.
    """
    GROOVY_SCRIPT_NAME = 'nexus3-cli-cleanup-policy'
    """Default Groovy script used by this class"""

    def __init__(self, client=None):
        self._client = client
        self._setup_script()

    @property
    def _script_name(self):
        return nexus_util.script_for_version(
            self.GROOVY_SCRIPT_NAME,
            self._client.server_version,
            GROOVY_SCRIPT_VERSIONS)

    def _setup_script(self):

        script_content = nexus_util.groovy_script(self._script_name)
        self._client.scripts.create_if_missing(
            self._script_name, script_content)

    def create_or_update(self, cleanup_policy):
        """
        Creates the given Cleanup Policy in the Nexus repository. If a policy
        with the same name already exists, it will be updated.

        :param cleanup_policy: the policy to create or update.
        :type cleanup_policy: CleanupPolicy
        :raises exception.NexusClientCreateCleanupPolicyError: when the Nexus
            API returns an error or unexpected result.
        """
        if not isinstance(cleanup_policy, CleanupPolicy):
            raise TypeError(
                f'cleanup_policy ({type(cleanup_policy)}) must be a '
                f'CleanupPolicy')

        script_args = json.dumps(cleanup_policy.configuration)
        try:
            LOG.debug('Create/update cleanup policy: %s', script_args)
            response = self._client.scripts.run(
                self._script_name, data=script_args)
        except exception.NexusClientAPIError:
            raise exception.NexusClientCreateCleanupPolicyError(
                cleanup_policy.configuration['name'])

        result = json.loads(response['result'])
        if result['name'] != cleanup_policy.configuration['name']:
            raise exception.NexusClientCreateCleanupPolicyError(response)

    def get_by_name(self, name):
        """
        Get a Nexus 3 cleanup policy by its name.

        :param name: name of the wanted policy
        :type name: str
        :return: the requested object
        :rtype: CleanupPolicy
        :raise exception.NexusClientInvalidRepository: when a repository with
            the given name isn't found.
        """
        script_args = json.dumps({'name': name})

        try:
            response = self._client.scripts.run(
                self._script_name, data=script_args)
        except exception.NexusClientAPIError:
            raise exception.NexusClientInvalidCleanupPolicy(name)

        cleanup_policy = json.loads(response['result'])

        return CleanupPolicy(self._client, **cleanup_policy)

    def list(self):
        """
        Return all cleanup policies.

        :return: every policy as a list of
            :class:`~nexuscli.api.cleanup_policy.model.CleanupPolicy`
            instances.
        :rtype: list[CleanupPolicy]
        """
        response = self._client.scripts.run(self._script_name, data={})

        cleanup_policies = json.loads(response['result'])

        return [CleanupPolicy(self._client, **c) for c in cleanup_policies]



# -*- coding: utf-8 -*-
import logging
import os

LOG_LEVEL = os.environ.get('LOG_LEVEL', 'WARNING').upper()

ROOT_LOGGER = logging.getLogger(__name__)
ROOT_LOGGER.setLevel(LOG_LEVEL)
ROOT_LOGGER.addHandler(logging.NullHandler())



import json
import logging
import os
import pathlib
import requests
import semver
import sys
from clint.textui import progress
from urllib.parse import urljoin

from nexuscli.nexus_config import NexusConfig
from nexuscli import exception, nexus_util
from nexuscli.api.cleanup_policy import CleanupPolicyCollection
from nexuscli.api.repository import validations, RepositoryCollection
from nexuscli.api.script import ScriptCollection

LOG = logging.getLogger(__name__)


class NexusClient(object):
    """
    A class to interact with Nexus 3's API.

    Unless all keyword arguments ``url``, ``user`` and ``password`` are
    supplied, the class will attempt to read the configuration file and,
    if unsuccessful, use defaults.

    Args:
        config (NexusConfig): instance containing the configuration for the
            Nexus service used by this instance.
    """
    def __init__(self, config=None):
        self.config = config or NexusConfig()
        self._local_sep = os.sep
        self._remote_sep = validations.REMOTE_PATH_SEPARATOR
        self._server_version = None
        self._cleanup_policies = None
        self._repositories = None
        self._scripts = None

        self.repositories.refresh()

    @property
    def server_version(self):
        """
        Parse the Server header from a Nexus request response and return
        as version information. The method expects the header Server to be
        present and formatted as, e.g., 'Nexus/3.19.1-01 (OSS)'

        :return: the parsed version. If it can't be determined, return None.
        :rtype: Union[None,semver.VersionInfo]
        """
        if self._server_version is None:
            response = self.http_get(self.config.url)

            if response.status_code != 200:
                raise exception.NexusClientAPIError(response.reason)

            server = response.headers.get('Server')

            if server is None:
                return None

            try:
                maybe_semver = server.split(' ')[0].split('/')[1].split('-')[0]
                version = semver.parse_version_info(maybe_semver)
            except (IndexError, ValueError):
                return None

            self._server_version = version
        return self._server_version

    @property
    def repositories(self):
        """
        Instance of
        :class:`~nexuscli.api.repository.collection.RepositoryCollection`. This
        will automatically use the existing instance of :class:`NexusClient` to
        communicate with the Nexus service.

        :rtype: RepositoryCollection
        """
        if self._repositories is None:
            self._repositories = RepositoryCollection(client=self)
        return self._repositories

    @property
    def cleanup_policies(self):
        """
        Instance of
        :class:`~nexuscli.api.cleanup_policy.collection.CleanupPolicyCollection`
        . This will automatically use the existing instance of
        :class:`NexusClient` to communicate with the Nexus service.
        """
        if self._cleanup_policies is None:
            self._cleanup_policies = CleanupPolicyCollection(client=self)
        return self._cleanup_policies

    @property
    def scripts(self):
        """
        Instance of
        :class:`~nexuscli.api.script.model.ScriptCollection`. This will
        automatically use the existing instance of :class:`NexusClient` to
        communicate with the Nexus service.
        """
        if self._scripts is None:
            self._scripts = ScriptCollection(client=self)
        return self._scripts

    @property
    def rest_url(self):
        """
        Full URL to the Nexus REST API, based on the ``url`` and ``version``
        from :attr:`config`.

        :rtype: str
        """
        url = urljoin(self.config.url, 'service/rest/')
        return urljoin(url, self.config.api_version + '/')

    def http_request(self, method, endpoint, service_url=None, **kwargs):
        """
        Performs a HTTP request to the Nexus REST API on the specified
        endpoint.

        :param method: one of ``get``, ``put``, ``post``, ``delete``.
        :type endpoint: str
        :param endpoint: URI path to be appended to the service URL.
        :type endpoint: str
        :param service_url: override the default URL to use for the request,
            which is created by joining :attr:`rest_url` and ``endpoint``.
        :type service_url: str
        :param kwargs: as per :py:func:`requests.request`.
        :rtype: requests.Response
        """
        service_url = service_url or self.rest_url
        url = urljoin(service_url, endpoint)

        try:
            response = requests.request(
                method=method, auth=self.config.auth, url=url,
                verify=self.config.x509_verify, **kwargs)
        except requests.exceptions.ConnectionError as e:
            raise exception.NexusClientConnectionError(str(e)) from None

        if response.status_code == 401:
            raise exception.NexusClientInvalidCredentials(
                'Try running `nexus3 login`')

        return response

    def http_get(self, endpoint):
        """
        Performs a HTTP GET request on the given endpoint.

        :param endpoint: name of the Nexus REST API endpoint.
        :type endpoint: str
        :rtype: requests.Response
        """
        return self.http_request('get', endpoint, stream=True)

    def http_head(self, endpoint):
        """
        Performs a HTTP HEAD request on the given endpoint.

        :param endpoint: name of the Nexus REST API endpoint.
        :type endpoint: str
        :rtype: requests.Response
        """
        return self.http_request('head', endpoint)

    def _get_paginated(self, endpoint, **request_kwargs):
        """
        Performs a GET request using the given args and kwargs. If the response
        is paginated, the method will repeat the request, manipulating the
        `params` keyword argument each time in order to receive all pages of
        the response.

        Items in the responses are sent in "batches": when all elements of a
        response have been yielded, a new request is made and the process
        repeated.

        :param request_kwargs: passed verbatim to the _request() method, except
            for the argument needed to paginate requests.
        :return: a generator that yields on response item at a time.
        :rtype: typing.Iterator[dict]
        """
        response = self.http_request('get', endpoint, **request_kwargs)
        if response.status_code == 404:
            raise exception.NexusClientAPIError(response.reason)

        try:
            content = response.json()
        except json.decoder.JSONDecodeError:
            raise exception.NexusClientAPIError(response.content)

        while True:
            for item in content.get('items'):
                yield item

            continuation_token = content.get('continuationToken')
            if continuation_token is None:
                break

            request_kwargs['params'].update(
                {'continuationToken': continuation_token})
            response = self.http_request('get', endpoint, **request_kwargs)

            try:
                content = response.json()
            except json.decoder.JSONDecodeError:
                raise exception.NexusClientAPIError(response.content)

    def http_post(self, endpoint, **kwargs):
        """
        Performs a HTTP POST request on the given endpoint.

        :param endpoint: name of the Nexus REST API endpoint.
        :type endpoint: str
        :param kwargs: as per :py:func:`requests.request`.
        :rtype: requests.Response
        """
        return self.http_request('post', endpoint, **kwargs)

    def http_put(self, endpoint, **kwargs):
        """
        Performs a HTTP PUT request on the given endpoint.

        :param endpoint: name of the Nexus REST API endpoint.
        :type endpoint: str
        :param kwargs: as per :py:func:`requests.request`.
        :rtype: requests.Response
        """
        return self.http_request('put', endpoint, **kwargs)

    def http_delete(self, endpoint, **kwargs):
        """
        Performs a HTTP DELETE request on the given endpoint.

        :param endpoint: name of the Nexus REST API endpoint.
        :type endpoint: str
        :param kwargs: as per :py:func:`requests.request`.
        :rtype: requests.Response
        """
        return self.http_request('delete', endpoint, **kwargs)

    def list(self, repository_path):
        """
        List all the artefacts, recursively, in a given ``repository_path``.

        :param repository_path: location on the repository service.
        :type repository_path: str
        :return: artefacts under ``repository_path``.
        :rtype: typing.Iterator[str]
        """
        for artefact in self.list_raw(repository_path):
            yield artefact.get('path')

    def _list_raw_search(self, repository_name, path_filter, partial_match):
        # TODO: use `group` attribute in raw repositories to speed-up queries
        query = {
            'repository': repository_name,
        }

        if path_filter:
            query['keyword'] = f'"{path_filter}"'  # hacky as fuck :(

        raw_response = self._get_paginated('search/assets', params=query)

        # TODO: maybe this filter is no longer needed due to keyword use ^
        return nexus_util.filtered_list_gen(
            raw_response, term=path_filter, partial_match=partial_match)

    def list_raw(self, repository_path):
        """
        As per :meth:`list` but yields raw Nexus artefacts as dicts.

        :param repository_path: location on the repository service.
        :type repository_path: str
        :rtype: typing.Iterator[dict]
        """
        repo, directory, filename = self.split_component_path(repository_path)
        path_filter = ''  # matches everything
        partial_match = True

        if directory is not None:
            path_filter = directory
            # Not all repos require a directory as part of the artefact path.
            if not (path_filter == '' or
                    path_filter.endswith(self._remote_sep)):
                path_filter += self._remote_sep

        if filename is not None:
            partial_match = False
            # The artefact path is always relative to the given repo.
            path_filter += filename

        list_gen = self._list_raw_search(repo, path_filter, partial_match)

        for artefact in list_gen:
            yield artefact

    def _pop_repository(self, component_path):
        """
        Helper for split_component_path. Returns the repository and the
        remainder of the component_path as a path_fragments list.

        :param component_path: the component path, as given to
            split_component_path.
        :return: tuple of (repository, path_fragments)
        :rtype: tuple(str, list)
        """
        path_fragments = component_path.split(self._remote_sep)
        try:
            repository = path_fragments.pop(0)
            # no cheating!
            if not repository or repository == '.':
                raise IndexError
        except IndexError:
            raise exception.NexusClientInvalidRepositoryPath(
                'The given path does not contain a repository: {}'.format(
                    component_path))

        return repository, path_fragments

    def _pop_filename(self, component_path, path_fragments):
        """
        Helper for split_component_path. Returns the filename.

        :param component_path: the component path, as given to
            split_component_path.
        :param path_fragments: as returned by _pop_repository.
        :return: filename or None, if not available.
        :rtype: str
        """
        filename = None
        try:
            if not component_path.endswith(self._remote_sep):
                filename = path_fragments.pop()
                if not filename or filename == '.':
                    raise IndexError
        except IndexError:
            return None

        return filename

    def _pop_directory(self, path_fragments):
        """
        Helper for split_component_path. Returns the directory.

        :param path_fragments: as returned by _pop_repository.
        :return: directory or None, if not available.
        :rtype: str
        """
        directory = self._remote_sep.join(path_fragments)
        # for consistency
        if directory.endswith(self._remote_sep):
            directory = directory[:-1]
        # nice try, user but no cigar
        if not directory or directory == '.':
            directory = None

        return directory

    def split_component_path(self, component_path):
        """
        Splits a given component path into repository, directory, filename.

        A Nexus component path for a raw directory must have this format:

        ``repository_name/directory[(/subdir1)...][/|filename]``

        A path ending in ``/`` represents a directory; otherwise it represents
        a filename.

            >>> dst0 = 'myrepo0/dir/'
            >>> dst1 = 'myrepo1/dir/subdir/'
            >>> dst2 = 'myrepo2/dir/subdir/file'
            >>> dst3 = 'myrepo3/dir/subdir/etc/file.ext'
            >>> split_component_path(dst0)
            >>> ('myrepo0', 'dir', None)
            >>> split_component_path(dst1)
            >>> ('myrepo1', 'dir/subdir', None)
            >>> split_component_path(dst2)
            >>> ('myrepo2', 'dir/subdir', 'file')
            >>> split_component_path(dst3)
            >>> ('myrepo3', 'dir/subdir/etc', 'file.ext')

        :param component_path: the Nexus component path, as described above.
        :type component_path: str
        :return: tuple of ``(repository_name, directory, filename)``. If the
            given ``component_path`` doesn't represent a file, then the
            ``filename`` is set to :py:obj:`None`.
        :rtype: tuple[str, str, str]
        """
        repository, path_fragments = self._pop_repository(component_path)
        filename = self._pop_filename(component_path, path_fragments)
        directory = self._pop_directory(path_fragments)

        return repository, directory, filename

    def _upload_dir_or_file(self, file_or_dir, dst_repo, dst_dir, dst_file,
                            **kwargs):
        """
        Helper for self.upload() to call the correct upload method according to
        the source given by the user.

        :param file_or_dir: location or file or directory to be uploaded.
        :param dst_repo: destination repository in Nexus.
        :param dst_dir: destination directory in dst_repo.
        :param dst_file: destination file name.
        :return: number of files uploaded.
        """
        repository = self.repositories.get_by_name(dst_repo)

        if os.path.isdir(file_or_dir):
            src_file = file_or_dir
            if dst_file is not None:
                raise exception.NexusClientInvalidRepositoryPath(
                    'Not allowed to upload a directory to a file')

            return repository.upload_directory(src_file, dst_dir, **kwargs)

        src_dir = file_or_dir
        repository.upload_file(src_dir, dst_dir, dst_file)
        return 1

    def upload(self, source, destination, recurse=True, flatten=False):
        """
        Process an upload. The source must be either a local file name or
        directory. The flatten and recurse options are honoured for
        directory uploads.

        The destination must be a valid Nexus 3 repository path, including the
        repository name as the first component of the path.

        :param source: location of file or directory to be uploaded.
        :type source: str
        :param destination: destination path in Nexus, including repository
            name and, if required, directory name (e.g. raw repos require a
            directory).
        :type destination: str
        :param recurse: do not process sub directories for uploads to remote
        :type recurse: bool
        :param flatten: Flatten directory structure by not reproducing local
                        directory structure remotely
        :type flatten: bool
        :return: number of files uploaded.
        """
        repo, directory, filename = self.split_component_path(destination)
        upload_count = self._upload_dir_or_file(
            source, repo, directory, filename,
            recurse=recurse, flatten=flatten)

        return upload_count

    def _remote_path_to_local(
            self, remote_src, local_dst, flatten, create=True):
        """
        Takes the remote path of an asset (without the repository name), the
        desired destination in the local file system, and creates the fully
        qualified path according to the instance settings.

        If self.flatten is True, the remote_path isn't reproduced locally.

        If the remote is a directory, we'll always assume the destination is
        also a directory, even if it doesn't end with a /.

        :param remote_src: path to the artefact as reported by the artefact
            service (i.e.: the `path` attribute of an asset object).
        :param local_dst: desired location in the local filesystem for the
            remote_path.
        :param create: whether or not to create the local destination file or
            directory.
        :return: the local path to be used.
        """
        # FIXME: use of multiple .. in the local_dst isn't resolved correctly
        remote_isdir = remote_src.endswith(self._remote_sep)
        # force destination to be a directory if the remote is a directory
        destination_isdir = (remote_isdir or
                             local_dst.endswith('.') or
                             local_dst.endswith('..') or
                             local_dst.endswith(self._local_sep))
        local_relative = remote_src.replace(self._remote_sep, self._local_sep)
        if flatten:
            local_relative = os.path.basename(local_relative)
        # remote=file, destination=file
        if not (remote_isdir or destination_isdir):
            # if files are given, rename the source to match destination
            local_relative_dir = os.path.dirname(local_relative)
            dst_file_name = os.path.basename(local_dst)
            local_dst = os.path.dirname(local_dst)
            if flatten:
                local_relative = dst_file_name
            else:
                local_relative = os.path.join(
                    local_relative_dir, dst_file_name)

        destination_path = pathlib.Path(local_dst)
        local_path = destination_path.joinpath(local_relative)

        if create:
            nexus_util.ensure_exists(local_path, is_dir=remote_isdir)

        return local_path.absolute()

    @staticmethod
    def _should_skip_download(download_url, download_path, artefact, nocache):
        """False when nocache is set or local file is out-of-date"""
        if nocache:
            try:
                LOG.debug('Removing {} because nocache is set\n'.format(
                    download_path))
                os.remove(download_path)
            except FileNotFoundError:
                pass
            return False

        if nexus_util.has_same_hash(artefact, download_path):
            LOG.debug(f'Skipping {download_url} because local copy '
                      f'{download_path} is up-to-date\n')
            return True

        return False

    def download_file(self, download_url, destination):
        """Download an asset from Nexus artefact repository to local
        file system.

        :param download_url: fully-qualified URL to asset being downloaded.
        :type download_url: str
        :param destination: file or directory location to save downloaded
            asset. Must be an existing directory; any exiting file in this
            location will be overwritten.
        :type destination: str
        :return:
        """
        response = self.http_get(download_url)

        if response.status_code != 200:
            sys.stderr.write(response.__dict__)
            raise exception.DownloadError(
                f'Downloading from {download_url}. '
                f'Reason: {response.reason}')

        with open(destination, 'wb') as fd:
            LOG.debug('Writing %s to %s', download_url, destination)
            for chunk in response.iter_content(chunk_size=8192):
                fd.write(chunk)

    def download(self, source, destination, flatten=False, nocache=False):
        """Process a download. The source must be a valid Nexus 3
        repository path, including the repository name as the first component
        of the path.

        The destination must be a local file name or directory.

        If a file name is given as destination, the asset may be renamed. The
        final destination will depend on ``flatten``.

        :param source: location of artefact or directory on the repository
            service.
        :type source: str
        :param destination: path to the local file or directory.
        :type destination: str
        :param flatten: if True, the remote path isn't reproduced locally.
        :type flatten: bool
        :param nocache: if True, force download of a directory or artefact,
                        ignoring an existing local copy. If false, it will not
                        re-download an existing copy if its checksum matches
                        the one in Nexus (as determined by
                        :meth:`nexuscli.nexus_util.has_same_hash`).
        :type nocache: bool
        :return: number of downloaded files.
        :rtype: int
        """
        download_count = 0
        if source.endswith(self._remote_sep) and \
                not (destination.endswith('.') or destination.endswith('..')):
            destination += self._local_sep

        artefacts = self.list_raw(source)

        artefacts = progress.bar(
                [a for a in artefacts], label='Downloading')

        for artefact in artefacts:
            download_url = artefact['downloadUrl']
            artefact_path = artefact['path']
            download_path = self._remote_path_to_local(
                artefact_path, destination, flatten)

            if self._should_skip_download(
                    download_url, download_path, artefact, nocache):
                download_count += 1
                continue

            try:
                self.download_file(download_url, download_path)
                download_count += 1
            except exception.DownloadError:
                LOG.warning('Error downloading %s', download_url)
                continue

        return download_count

    def delete(self, repository_path):
        """
        Delete artefacts, recursively if ``repository_path`` is a directory.

        :param repository_path: location on the repository service.
        :type repository_path: str
        :return: number of deleted files. Negative number for errors.
        :rtype: int
        """

        delete_count = 0
        death_row = self.list_raw(repository_path)

        death_row = progress.bar([a for a in death_row], label='Deleting')

        for artefact in death_row:
            id_ = artefact['id']
            artefact_path = artefact['path']

            response = self.http_delete(f'assets/{id_}')
            LOG.info('Deleted: %s (%s)', artefact_path, id_)
            delete_count += 1
            if response.status_code == 404:
                LOG.warning('File disappeared while deleting')
                LOG.debug(response.reason)
            elif response.status_code != 204:
                LOG.error(response.reason)
                return -1

        return delete_count



from texttable import Texttable

from nexuscli import exception
from nexuscli.api import cleanup_policy
from nexuscli.cli import constants


def cmd_list(nexus_client):
    """Performs ``nexus3 cleanup_policy list``"""
    policies = nexus_client.cleanup_policies.list()
    if len(policies) == 0:
        return exception.CliReturnCode.POLICY_NOT_FOUND.value

    table = Texttable(max_width=constants.TTY_MAX_WIDTH)
    table.add_row(
        ['Name', 'Format', 'Downloaded', 'Updated', 'Regex'])
    table.set_deco(Texttable.HEADER)
    for policy in policies:
        p = policy.configuration
        table.add_row([
            p['name'], p['format'],
            p['criteria'].get('lastDownloaded', 'null'),
            p['criteria'].get('lastBlobUpdated', 'null'),
            p['criteria'].get('regex', 'null')],
        )

    print(table.draw())
    return exception.CliReturnCode.SUCCESS.value


def cmd_create(nexus_client, **kwargs):
    """Performs ``nexus3 cleanup_policy create``"""
    policy = cleanup_policy.CleanupPolicy(None, **kwargs)
    nexus_client.cleanup_policies.create_or_update(policy)

    return exception.CliReturnCode.SUCCESS.value



REMOTE_PATH_SEPARATOR = '/'


def ensure_known(target, value, known):
    """
    Validate whether the a target argument is known and supported. The
    ``target`` is only used to provide a friendlier message to the user.
    The given ``value`` is checked against ``known`` and ``supported``.

    Args:
        target (str): name of the target, as known to the end-user.
        value (str): value of the target key.
        known (list,tuple): known possible values for the target.

    Raises:
          :class:`ValueError`: if given value is not in ``known``.
    """
    if value not in known:
        raise ValueError(f'{target}={value} must be one of {known}')



from nexuscli import exception, nexus_util


class ScriptCollection(object):
    """
    A class to manage Nexus 3 scripts.

    Args:
        client(nexuscli.nexus_client.NexusClient): the client instance that
            will be used to perform operations against the Nexus 3 service. You
            must provide this at instantiation or set it before calling any
            methods that require connectivity to Nexus.

    Attributes:
        client(nexuscli.nexus_client.NexusClient): as per ``client``
            argument of :class:`ScriptCollection`.
    """
    def __init__(self, client=None):
        self._client = client

    def exists(self, name):
        """
        Check if a script exists.

        :param name: of script to verify existence.
        :return: True if it exists, false otherwise
        :rtype: bool
        :raises exception.NexusClientAPIError: if the response from the Nexus
            service isn't recognised; i.e.: any HTTP code other than 200, 404.
        """
        resp = self._client.http_head(f'script/{name}')
        if resp.status_code == 200:
            return True
        elif resp.status_code == 404:
            return False
        else:
            raise exception.NexusClientAPIError(resp.content)

    def get(self, name):
        """
        Get a Nexus 3 script by name.

        :param name: of script to be retrieved.
        :return: the script or None, if not found
        :rtype: dict, None
        :raises exception.NexusClientAPIError: if the response from the Nexus
            service isn't recognised; i.e.: any HTTP code other than 200, 404.
        """
        resp = self._client.http_get(f'script/{name}')
        if resp.status_code == 200:
            return resp.json()
        elif resp.status_code == 404:
            return None
        else:
            raise exception.NexusClientAPIError(resp.content)

    def list(self):
        """
        List of all script names on the Nexus 3 service.

        :return: a list of names
        :rtype: list
        :raises exception.NexusClientAPIError: if the script names cannot be
            retrieved; i.e.: any HTTP code other than 200.
        """
        resp = self._client.http_get('script')
        if resp.status_code != 200:
            raise exception.NexusClientAPIError(resp.content)

        return resp.json()

    def create_if_missing(self, name, content=None, script_type='groovy'):
        """
        Creates a script in the Nexus 3 service IFF a script with the same name
        doesn't exist. Equivalent to checking if the script exists with
        :meth:`get` and, if not, creating it with :meth:`create`.

        :param name: name of script to be created.
        :type name: str
        :param content: script code. If not given, the method will use
            :py:meth:`nexuscli.nexus_util.groovy_script` to read the script
            code from a local file.
        :type content: Union[str,NoneType]
        :param script_type: type of script to be created.
        :type script_type: str
        :raises exception.NexusClientAPIError: if the script creation isn't
            successful; i.e.: any HTTP code other than 204.
        """
        content = content or nexus_util.groovy_script(name)

        if not self.exists(name):
            self.create(name, content, script_type)

    def create(self, script_name, script_content, script_type='groovy'):
        """
        Create the given script in the Nexus 3 service.

        :param script_name: name of script to be created.
        :type script_name: str
        :param script_content: script code.
        :type script_content: str
        :param script_type: type of script to be created.
        :type script_type: str
        :raises exception.NexusClientAPIError: if the script creation isn't
            successful; i.e.: any HTTP code other than 204.
        """
        script = {
            'type': script_type,
            'name': script_name,
            'content': script_content,
        }

        resp = self._client.http_post('script', json=script)
        if resp.status_code != 204:
            raise exception.NexusClientAPIError(resp.content)

    def run(self, script_name, data=''):
        """
        Runs an existing script on the Nexus 3 service.

        :param script_name: name of script to be run.
        :param data: parameters to be passed to the script, via HTTP POST. If
            the script being run requires a certain format or encoding, you
            need to prepare it yourself. Typically this is `json.dumps(data)`.
        :return: the content returned by the script, if any.
        :rtype: str, dict
        :raises exception.NexusClientAPIError: if the Nexus service fails to
            run the script; i.e.: any HTTP code other than 200.
        """
        headers = {'content-type': 'text/plain'}
        endpoint = 'script/{}/run'.format(script_name)
        resp = self._client.http_post(endpoint, headers=headers, data=data)
        if resp.status_code != 200:
            raise exception.NexusClientAPIError(resp.content)

        return resp.json()

    def delete(self, script_name):
        """
        Deletes a script from the Nexus 3 repository.

        :param script_name: name of script to be deleted.
        :raises exception.NexusClientAPIError: if the Nexus service fails to
            delete the script; i.e.: any HTTP code other than 204.
        """
        endpoint = 'script/{}'.format(script_name)
        resp = self._client.http_delete(endpoint)
        if resp.status_code != 204:
            raise exception.NexusClientAPIError(resp.reason)


# TODO: describe script and use/return from collection methods
class Script(object):
    """A Class representing a Nexus 3 script."""
    pass



# -*- coding: utf-8 -*-
import io
from setuptools import find_packages, setup

package_name = 'nexus3-cli'
package_version = '3.1.0'

requires = [
    'click>=7.0.0,<8',
    'clint',
    'future',
    'faker',
    'inflect',
    'future',
    'requests[security]>=2.14.2',
    'semver',
    'texttable'
]

test_requires = [
    'codecov',
    'flake8',
    'pytest',
    'pytest-cov',
    'pytest-helpers-namespace',
    'pytest-mock',
    'pytest-faker',
]

with io.open('README.md', mode='r', encoding='utf-8') as f:
    readme = f.read()

setup(
    author='Thiago Figueiró',
    name=package_name,
    version=package_version,
    description='A python-based CLI for Sonatype Nexus OSS 3',
    url='https://gitlab.com/thiagocsf/nexus3-cli',
    long_description=readme,
    long_description_content_type="text/markdown",
    setup_requires=["pytest-runner"],
    install_requires=requires,
    tests_require=test_requires,
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6,<4',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        'nexuscli': ['script/groovy/'],
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'nexus3=nexuscli.cli:nexus_cli',
        ],
    },
    extras_require={'test': test_requires},
)



import os
import semver
from clint.textui import progress
from urllib.parse import urlparse

from nexuscli.api.repository import validations, util, upload

DEFAULT_RECIPE = 'raw'
DEFAULT_WRITE_POLICY = 'ALLOW'
DEFAULT_BLOB_STORE_NAME = 'default'
DEFAULT_STRICT_CONTENT = False

# https://issues.sonatype.org/browse/NEXUS-19525
# https://github.com/thiagofigueiro/nexus3-cli/issues/77
CLEANUP_SET_MIN_VERSION = semver.VersionInfo(3, 19, 0)


class Repository:
    """
    A base Nexus repository.

    Nexus 3 repository recipes (formats) supported by this class:

        - `bower
          <https://help.sonatype.com/repomanager3/formats/bower-repositories>`_
        - `npm
          <https://help.sonatype.com/repomanager3/formats/npm-registry>`_
        - `nuget
          <https://help.sonatype.com/repomanager3/formats/nuget-repositories>`_
        - `pypi
          <https://help.sonatype.com/repomanager3/formats/pypi-repositories>`_
        - `raw
          <https://help.sonatype.com/repomanager3/formats/raw-repositories>`_
        - `rubygems
          <https://help.sonatype.com/repomanager3/formats/rubygems-repositories>`_
        - `docker
          <https://help.sonatype.com/repomanager3/formats/docker-registry>`_
        - `apt
          <https://help.sonatype.com/repomanager3/formats/apt-repositories>`_
    :param name: name of the repository.
    :type name: str
    :param nexus_client: the :class:`~nexuscli.nexus_client.NexusClient`
        instance that will be used to perform operations against the Nexus 3
        service. You must provide this at instantiation or set it before
        calling any methods that require connectivity to Nexus.
    :type nexus_client: nexuscli.nexus_client.NexusClient
    :param recipe: format (recipe) of the new repository. Must be one of
        :py:attr:`RECIPES`. See Nexus documentation for details.
    :type recipe: str
    :param blob_store_name: name of an existing blob store; 'default'
        should work on most installations.
    :type blob_store_name: str
    :param strict_content_type_validation: Whether to validate file
        extension against its content type.
    :type strict_content_type_validation: bool
    :param cleanup_policy: name of an existing repository clean-up policy.
    :type cleanup_policy: str
    """

    RECIPES = (
        'bower',
        'npm',
        'nuget',
        'pypi',
        'raw',
        'rubygems',
    )
    TYPE = None

    def __init__(self, name,
                 nexus_client=None,
                 recipe=DEFAULT_RECIPE,
                 blob_store_name=DEFAULT_BLOB_STORE_NAME,
                 strict_content_type_validation=DEFAULT_STRICT_CONTENT,
                 cleanup_policy=None
                 ):
        self.name = name
        self.nexus_client = nexus_client
        # TODO: remove this the RECIPES attributes; no longer needed as there's
        #   a unique class for each recipe/type.
        self.recipe = recipe.lower()
        self.blob_store_name = blob_store_name
        self.strict_content = strict_content_type_validation
        self._cleanup_policy = cleanup_policy

        self.__validate_params()

    def __repr__(self):
        return f'{self.__class__.__name__}-{self.name}-{self.recipe}'

    def __validate_params(self):
        validations.ensure_known('recipe', self.recipe, self.RECIPES)

    @property
    def recipe_name(self):
        """
        The Nexus 3 name for this repository's recipe (format). This is almost
        always the same as :attr:`name` with ``maven`` being the notable
        exception.
        """
        if self.recipe == 'maven':
            return 'maven2'

        return self.recipe

    def _cleanup_uses_set(self):
        # In case Sonatype changes the version string format, default to the
        # new behaviour as there should be more people using newer versions
        if self.nexus_client.server_version is None:
            return True

        # When the breaking API change was introduced
        if self.nexus_client.server_version >= CLEANUP_SET_MIN_VERSION:
            return True

        return False

    @property
    def cleanup_policy(self):
        """
        Groovy-formatted value for the cleanup/policy attribute.
        """
        if self._cleanup_uses_set():
            return [self._cleanup_policy]
        else:
            return self._cleanup_policy

    @property
    def configuration(self):
        """
        Repository configuration represented as a python dict. The dict
        returned by this property can be converted to JSON for use with the
        ``nexus3-cli-repository-create``
        groovy script created by the
        :py:meth:`~nexuscli.api.repository.collection.RepositoryCollection.create`
        method.

        Example structure and attributes common to all repositories:

        >>> common_configuration = {
        >>>     'name': 'my-repository',
        >>>     'online': True,
        >>>     'recipeName': 'raw',
        >>>     '_state': 'present',
        >>>     'attributes': {
        >>>         'storage': {
        >>>             'blobStoreName': 'default',
        >>>         },
        >>>         'cleanup': {
        >>>             'policyName': None,
        >>>         }
        >>>     }
        >>> }

        Depending on the repository type and format (recipe), other attributes
        will be present.

        :return: repository configuration
        :rtype: dict
        """
        repo_config = {
            'name': self.name,
            'online': True,
            'recipeName': f'{self.recipe_name}-{self.TYPE}',
            '_state': 'present',
            'attributes': {
                'storage': {
                    'blobStoreName': self.blob_store_name,
                    'strictContentTypeValidation': self.strict_content,
                },
            }
        }

        # we want 'x' or ['x'] but not None or [None]
        if self.cleanup_policy and any(self.cleanup_policy):
            repo_config['attributes']['cleanup'] = {
                'policyName': self.cleanup_policy}

        return repo_config


class ProxyRepository(Repository):
    """
    A proxy Nexus repository.

    :param name: name of the repository.
    :type name: str
    :param remote_url: The URL of the repository being proxied, including the
        protocol scheme.
    :type remote_url: str
    :param auto_block: Auto-block outbound connections on the repository if
        remote peer is detected as unreachable/unresponsive.
    :type auto_block: bool
    :param content_max_age: How long (in minutes) to cache artifacts before
        rechecking the remote repository. Release repositories should use -1.
    :type content_max_age: int
    :param metadata_max_age: How long (in minutes) to cache metadata before
        rechecking the remote repository.
    :type metadata_max_age: int
    :param negative_cache_enabled: Cache responses for content not present in
        the proxied repository
    :type negative_cache_enabled: bool
    :param negative_cache_ttl: How long to cache the fact that a file was not
        found in the repository (in minutes)
    :type negative_cache_ttl: int
    :param kwargs: see :class:`Repository`
    """

    TYPE = 'proxy'

    def __init__(self, name,
                 remote_url=None,
                 auto_block=True,
                 content_max_age=1440,
                 metadata_max_age=1440,
                 negative_cache_enabled=True,
                 negative_cache_ttl=1440,
                 remote_auth_type=None,
                 remote_username=None,
                 remote_password=None,
                 **kwargs):
        self.remote_url = remote_url
        self.auto_block = auto_block
        self.content_max_age = content_max_age
        self.metadata_max_age = metadata_max_age
        self.negative_cache_enabled = negative_cache_enabled
        self.negative_cache_ttl = negative_cache_ttl
        self.remote_username = remote_username
        self.remote_password = remote_password
        self.remote_auth_type = remote_auth_type

        super().__init__(name, **kwargs)

        self.__validate_params()

    def __validate_params(self):
        if not isinstance(self.remote_url, str):
            raise ValueError('remote_url must be a str')

        parsed_url = urlparse(self.remote_url)
        if not (parsed_url.scheme and parsed_url.netloc):
            raise ValueError(
                f'remote_url={self.remote_url} is not a valid URL')

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration

        repo_config['attributes'].update({
            'httpclient': {
                'blocked': False,
                'autoBlock': self.auto_block,
            },
            'proxy': {
                'remoteUrl': self.remote_url,
                'contentMaxAge': self.content_max_age,
                'metadataMaxAge': self.metadata_max_age,
            },
            'negativeCache': {
                'enabled': self.negative_cache_enabled,
                'timeToLive': self.negative_cache_ttl,
            },
        })

        if self.remote_auth_type == 'username':
            repo_config['attributes']['httpclient'].update({
                'authentication': {
                    'type': self.remote_auth_type,
                    'username': self.remote_username,
                    'password': self.remote_password
                }
            })
        return repo_config


class HostedRepository(Repository):
    """
    A hosted Nexus repository.

    :param name: name of the repository.
    :type name: str
    :param write_policy: one of :py:attr:`WRITE_POLICIES`. See Nexus
        documentation for details.
    :type write_policy: str
    :param kwargs: see :class:`Repository`
    """
    WRITE_POLICIES = ['ALLOW', 'ALLOW_ONCE', 'DENY']
    """Nexus 3 repository write policies supported by this class."""

    TYPE = 'hosted'

    def __init__(self, name, write_policy=DEFAULT_WRITE_POLICY, **kwargs):
        self.write_policy = write_policy

        super().__init__(name, **kwargs)

        self.__validate_params()

    def __validate_params(self):
        validations.ensure_known(
            'write_policy', self.write_policy, self.WRITE_POLICIES)

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration

        repo_config['attributes']['storage'].update({
            'writePolicy': self.write_policy,
            'strictContentTypeValidation': self.strict_content,
        })

        return repo_config

    def upload_file(self, src_file, dst_dir, dst_file=None):
        """
        Uploads a singe file to the directory and file name specified.

        :param src_file: path to the local file to be uploaded.
        :param dst_dir: directory under dst_repo to place file in.
        :param dst_file: destination file name. If not given, the basename of
            ``src_file`` name is used.
        """
        if dst_file is None:
            dst_file = os.path.basename(src_file)

        upload_method_name = f'upload_file_{self.recipe_name}'
        try:
            # Find upload method in the upload module using naming convention
            upload_method = getattr(upload, upload_method_name)
        except AttributeError:
            raise NotImplementedError(upload_method_name) from None

        upload_method(self, src_file, dst_dir, dst_file)

    def upload_directory(self, src_dir, dst_dir, recurse=True, flatten=False):
        """
        Uploads all files in a directory to the specified destination directory
        in this repository, honouring options flatten and recurse.

        :param src_dir: path to local directory to be uploaded
        :param dst_dir: destination directory in dst_repo
        :param recurse: when True, upload directory recursively.
        :type recurse: bool
        :param flatten: when True, the source directory tree isn't replicated
            on the destination.
        :return: number of files uploaded
        :rtype: int
        """
        file_set = util.get_files(src_dir, recurse)
        file_count = len(file_set)
        file_set = progress.bar(file_set, expected_size=file_count)

        for relative_filepath in file_set:
            file_path = os.path.join(src_dir, relative_filepath)
            sub_directory = util.get_upload_subdirectory(
                dst_dir, file_path, flatten)
            self.upload_file(file_path, sub_directory)

        return file_count


class MavenRepository(Repository):
    """
    A base `Maven
    <https://help.sonatype.com/repomanager3/formats/maven-repositories#MavenRepositories-MavenRepositoryFormat>`_
    Nexus repository.

    :param name: name of the repository.
    :type name: str
    :param layout_policy: one of :py:attr:`LAYOUT_POLICIES`. See Nexus
        documentation for details.
    :param version_policy: one of :py:attr:`VERSION_POLICIES`. See Nexus
        documentation for details.
    :param kwargs: see :class:`Repository`
    """
    RECIPES = ('maven',)

    LAYOUT_POLICIES = ('PERMISSIVE', 'STRICT')
    """Maven layout policies"""

    VERSION_POLICIES = ('RELEASE', 'SNAPSHOT', 'MIXED')
    """Maven version policies"""

    def __init__(self, name,
                 layout_policy='PERMISSIVE',
                 version_policy='RELEASE',
                 **kwargs):
        self.layout_policy = layout_policy
        self.version_policy = version_policy

        kwargs.update({'recipe': 'maven'})

        super().__init__(name, **kwargs)

        self.__validate_params()

    def __validate_params(self):
        validations.ensure_known(
            'layout_policy', self.layout_policy, self.LAYOUT_POLICIES)
        validations.ensure_known(
            'version_policy', self.version_policy, self.VERSION_POLICIES)

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration

        repo_config['attributes']['maven'] = {
            'layoutPolicy': self.layout_policy,
            'versionPolicy': self.version_policy,
        }

        return repo_config


class MavenHostedRepository(HostedRepository, MavenRepository):
    """
    A `Maven
    <https://help.sonatype.com/repomanager3/formats/maven-repositories#MavenRepositories-MavenRepositoryFormat>`_
    hosted Nexus repository.

    See :class:`HostedRepository` and :class:`MavenRepository`
    """
    pass


class MavenProxyRepository(MavenRepository, ProxyRepository):
    """
    A `Maven
    <https://help.sonatype.com/repomanager3/formats/maven-repositories#MavenRepositories-MavenRepositoryFormat>`_
    proxy Nexus repository.

    See :class:`MavenRepository` and :class:`ProxyRepository`
    """
    pass


class YumRepository(Repository):
    """
    A `Yum <https://help.sonatype.com/repomanager3/formats/yum-repositories>`_
    base Nexus repository.

    :param name: name of the repository.
    :type name: str
    :param depth: The Yum ``repodata`` depth. Usually 1.
    :type depth: int
    :param kwargs: see :class:`Repository`
    """
    RECIPES = ('yum',)

    def __init__(self, name, depth=1, **kwargs):
        self.depth = depth

        kwargs.update({'recipe': 'yum'})

        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration
        repo_config['attributes']['yum'] = {
            'repodataDepth': self.depth
        }
        return repo_config


class YumHostedRepository(HostedRepository, YumRepository):
    """
    A `Yum <https://help.sonatype.com/repomanager3/formats/yum-repositories>`_
    hosted Nexus repository.

    See :class:`HostedRepository` and :class:`YumRepository`
    """
    pass


class YumProxyRepository(ProxyRepository, YumRepository):
    """
    A `Yum <https://help.sonatype.com/repomanager3/formats/yum-repositories>`_
    proxy Nexus repository.

    See :class:`ProxyRepository` and :class:`YumRepository`
    """
    pass


# For the convenience of not having to handle these recipe names differently
class BowerHostedRepository(HostedRepository):
    pass


class BowerProxyRepository(ProxyRepository):
    pass


class NpmHostedRepository(HostedRepository):
    pass


class NpmProxyRepository(ProxyRepository):
    pass


class NugetHostedRepository(HostedRepository):
    pass


class NugetProxyRepository(ProxyRepository):
    pass


class PypiHostedRepository(HostedRepository):
    pass


class PypiProxyRepository(ProxyRepository):
    pass


class RawHostedRepository(HostedRepository):
    pass


class RawProxyRepository(ProxyRepository):
    pass


class RubygemsHostedRepository(HostedRepository):
    pass


class RubygemsProxyRepository(ProxyRepository):
    pass


class DockerRepository(Repository):
    RECIPES = ('docker',)

    def __init__(self, name,
                 http_port=8084,
                 https_port=8085,
                 v1_enabled=False,
                 force_basic_auth=False,
                 **kwargs):
        self.https_port = https_port
        self.http_port = http_port
        self.v1_enabled = v1_enabled
        self.force_basic_auth = force_basic_auth
        kwargs.update({'recipe': 'docker'})
        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration

        repo_config['attributes'].update({
            'docker': {
                'httpPort': self.http_port,
                'httpsPort': self.https_port,
                'v1Enabled': self.v1_enabled,
                'forceBasicAuth': self.force_basic_auth
            }
        })

        return repo_config


class DockerHostedRepository(HostedRepository, DockerRepository):
    pass


class DockerProxyRepository(ProxyRepository, DockerRepository):
    INDEX_TYPES = ('REGISTRY', 'HUB', 'CUSTOM')

    def __init__(self, name,
                 index_type='REGISTRY',
                 use_trust_store_for_index_access=False,
                 index_url='https://index.docker.io/',
                 **kwargs):
        self.index_type = index_type

        validations.ensure_known(
            'index_type',
            self.index_type,
            self.INDEX_TYPES
        )

        self.use_trust_store_for_index_access =\
            use_trust_store_for_index_access
        self.index_url = index_url
        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        """
        As per :py:obj:`Repository.configuration` but specific to this
        repository recipe and type.

        :rtype: str
        """
        repo_config = super().configuration

        if self.index_type == 'REGISTRY':
            repo_config['attributes'].update({
                'dockerProxy': {
                    'indexType': self.index_type
                },
            })
        if self.index_type == 'HUB':
            repo_config['attributes'].update({
                'dockerProxy': {
                    'indexType': self.index_type,
                    "useTrustStoreForIndexAccess":
                        self.use_trust_store_for_index_access

                },
            })
        if self.index_type == 'CUSTOM':
            repo_config['attributes'].update({
                'dockerProxy': {
                    'indexType': self.index_type,
                    "useTrustStoreForIndexAccess":
                        self.use_trust_store_for_index_access,
                    "indexUrl": self.index_url,
                },
            })
        return repo_config


class AptRepository(Repository):
    RECIPES = ('apt',)

    def __init__(self, name,
                 distribution='bionic',
                 **kwargs):
        self.distribution = distribution
        kwargs.update({'recipe': 'apt'})
        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        repo_config = super().configuration

        repo_config['attributes'].update({
            'apt': {
                'distribution': self.distribution,
            }
        })

        return repo_config


class AptHostedRepository(AptRepository, HostedRepository):
    def __init__(self, name,
                 gpg_keypair: str = None,
                 passphrase=None,
                 **kwargs):
        self.gpg_keypair = gpg_keypair
        self.passphrase = passphrase
        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        repo_config = super().configuration
        repo_config['attributes'].update({
            'aptSigning': {
                'keypair': self.gpg_keypair,
                'passphrase': self.passphrase
            }
        })

        return repo_config


class AptProxyRepository(AptRepository, ProxyRepository):
    def __init__(self, name,
                 flat=False,
                 **kwargs):
        self.flat = flat
        super().__init__(name, **kwargs)

    @property
    def configuration(self):
        repo_config = super().configuration

        repo_config['attributes']['apt']['flat'] = self.flat

        return repo_config


# FIXME: these are supposed to be strings
__all__ = [
    Repository, HostedRepository, ProxyRepository,
    AptHostedRepository, AptProxyRepository,
    BowerHostedRepository, BowerProxyRepository,
    DockerHostedRepository, DockerProxyRepository,
    MavenHostedRepository, MavenProxyRepository,
    NpmHostedRepository, NpmProxyRepository,
    NugetHostedRepository, NugetProxyRepository,
    PypiHostedRepository, PypiProxyRepository,
    RawHostedRepository, RawProxyRepository,
    RubygemsHostedRepository, RubygemsProxyRepository,
    YumHostedRepository, YumProxyRepository,
]

SUPPORTED_FORMATS = sorted(
    set([recipe for cls in __all__ for recipe in cls.RECIPES]))



import json
import typing
from contextlib import contextmanager
from pathlib import Path


DEFAULT_CONFIG = str(Path.home().joinpath('.nexus-cli').absolute())

# when adding entries to DEFAULTS, check if to_dict() would still make sense
DEFAULTS = {
    'api_version': 'v1',
    'username': 'admin',
    'password': '',
    'url': 'http://localhost:8081',
    'x509_verify': True,
}


class NexusConfig:
    """
    A class to hold Nexus 3's configuration.

    Unless keyword arguments ``url``, ``user`` and ``password`` are
    supplied, the class will attempt to read the configuration file and,
    if unsuccessful, use defaults.

    Args:
        username (str): username for Nexus service at given url.
        password (str): password for username above.
        url (str): URL to Nexus 3 OSS service.
        x509_verify (bool): toggle certificate validation.
        api_version (str): Nexus REST API version to be used.
        config_path (str): local file containing configuration above in JSON
            format with these keys: ``nexus_url``, ``nexus_user``,
            ``nexus_pass`` and ``nexus_verify``.
    """
    def __init__(self,
                 username=DEFAULTS['username'],
                 password=DEFAULTS['password'],
                 url=DEFAULTS['url'],
                 x509_verify=DEFAULTS['x509_verify'],
                 api_version=DEFAULTS['api_version'],
                 config_path=None):

        self._api_version = api_version
        self._username = username
        self._password = password
        self._url = url
        if not self._url.endswith('/'):
            self._url += '/'
        self._x509_verify = x509_verify
        self._config_path = Path(config_path or DEFAULT_CONFIG)

    @property
    def to_dict(self):
        """
        Current instance configuration.

        :rtype: dict
        """
        config = {}
        for key in DEFAULTS.keys():
            config[key] = getattr(self, f'_{key}')
        return config

    @property
    def auth(self):
        """
        Current username and password as a tuple.

        :rtype: tuple[str, str]
        """
        return self._username, self._password

    @property
    def api_version(self):
        """
        Current API version in use.

        :rtype: str
        """
        return self._api_version

    @property
    def url(self):
        """
        The Nexus service URL

        :rtype: str
        """
        return self._url

    @property
    def x509_verify(self):
        """
        Whether to validate the x509 certificate when using https to
        access the Nexus service

        :rtype: str
        """
        return self._x509_verify

    @property
    def config_path(self) -> Path:
        """
        Path to configuration file, as given by ``config_path`` during
        instantiation.
        """
        return self._config_path

    # TODO: deprecate this; it's usually not needed and, if it is, one can always use
    #  str(config_path)
    @property
    def config_file(self):
        """
        Path to configuration file, as given by ``config_path`` during
        instantiation.

        :rtype: str
        """
        return str(self.config_path)

    @staticmethod
    @contextmanager
    def _open_dump(dst_path) -> typing.IO:
        dst_path.touch(mode=0o600)
        try:
            with dst_path.open(mode='w+', encoding='utf-8') as fh:
                yield fh
        finally:
            pass

    def dump_env(self):
        """
        Writes the current configuration to disk under property:`config_file`.env.

        If a file already exists, it will be overwritten. The permission will
        be set to read/write to the owner only.

        return: name of the file written
        """
        from nexuscli.cli.constants import ENV_VAR_PREFIX  # avoid circular dependency
        with self._open_dump(self.config_path.with_suffix('.env')) as fh:
            for k, v in sorted(self.to_dict.items()):
                fh.write(f'{ENV_VAR_PREFIX}_{k.upper()}={v}\n')
            return fh.name

    def dump(self) -> str:
        """
        Writes the current configuration to disk under property:`config_file`.

        If a file already exists, it will be overwritten. The permission will
        be set to read/write to the owner only.

        return: name of the file written
        """
        with self._open_dump(self.config_path) as fh:
            json.dump(self.to_dict, fh, ensure_ascii=False, indent=4, sort_keys=True)
            return fh.name

    def load(self):
        """
        Load the configuration settings from the file specified by
        :attr:`config_file`.

        The configuration file is in JSON format and expects these keys:
        ``nexus_user``, ``nexus_pass``, ``nexus_url``, ``nexus_verify``.
        """
        with self.config_path.open(mode='r', encoding='utf-8') as fh:
            config = json.load(fh)

        for key, default_value in DEFAULTS.items():
            setattr(self, f'_{key}', config.get(key, default_value))
